## This file is part of Jax Geometry
#
# Copyright (C) 2021, Stefan Sommer (sommer@di.ku.dk)
# https://bitbucket.org/stefansommer/jaxgeometry
#
# Jax Geometry is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Jax Geometry is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Jax Geometry. If not, see <http://www.gnu.org/licenses/>.
#


from jaxgeometry.setup import *
from jaxgeometry.params import *

from jaxgeometry.manifolds.ellipsoid import *

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import matplotlib.ticker as ticker

class H2(EmbeddedManifold):
    """ hyperbolic plane """

    def __init__(self,):
        F = lambda x: jnp.stack([jnp.cosh(x[0][0]),jnp.sinh(x[0][0])*jnp.cos(x[0][1]),jnp.sinh(x[0][0])*jnp.sin(x[0][1])])
        invF = lambda x: jnp.stack([jnp.arccosh(x[0][0]),jnp.arctan2(x[0][2],x[0][1])])
        self.do_chart_update = lambda x: False

        EmbeddedManifold.__init__(self,F,2,3,invF=invF)

        # metric matrix from embedding into Minkowski space
        self.g = lambda x: jnp.einsum('ji,j,jl',self.JF(x),np.array([-1.,1.,1.]),self.JF(x))


    def __str__(self):
        return "%dd dim hyperbolic space" % (self.dim,)

    def newfig(self):
        newfig3d()

    def plot(self,rotate=None,alpha=None,lw=0.3):
        ax = plt.gca()
        x = np.arange(-10,10,1)
        ax.xaxis._axinfo["grid"]['linewidth'] = lw
        ax.yaxis._axinfo["grid"]['linewidth'] = lw
        ax.zaxis._axinfo["grid"]['linewidth'] = lw
        ax.set_xlim(1.,2.)
        ax.set_ylim(-1.,1.)
        ax.set_zlim(-1.,1.)
        #ax.set_aspect("equal")
        if rotate is not None:
            ax.view_init(rotate[0],rotate[1])
    #     else:
    #         ax.view_init(35,225)
        plt.xlabel('x')
        plt.ylabel('y')
    
    #    ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        #draw ellipsoid
        u, v = np.mgrid[-1.5:1.5:20j, 0:2*np.pi:20j]
        x=np.cosh(u)
        y=np.sinh(u)*np.cos(v)
        z=np.sinh(u)*np.sin(v)
        ax.plot_wireframe(x, y, z, color='gray', alpha=0.5)
    
        if alpha is not None:
            ax.plot_surface(x, y, z, color=cm.jet(0.), alpha=alpha)


    def plot_field(self, field,lw=.3):
        ax = plt.gca(projection='3d')
        x = np.arange(-10,10,1)
        ax.xaxis._axinfo["grid"]['linewidth'] = lw
        ax.yaxis._axinfo["grid"]['linewidth'] = lw
        ax.zaxis._axinfo["grid"]['linewidth'] = lw
        ax.set_xlim(-1.,1.)
        ax.set_ylim(-1.,1.)
        ax.set_zlim(-1.,1.)
        #ax.set_aspect("equal")

        plt.xlabel('x')
        plt.ylabel('y')

#        ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        #draw ellipsoid
        u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
        x=np.cosh(u)
        y=np.sinh(u)*np.cos(v)
        z=np.sinh(u)*np.sin(v)
        
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                Fx = np.array([x[i,j],y[i,j],z[i,j]])
                chart = self.centered_chart(Fx)
                xcoord = self.invF((Fx,chart))
                v = field((xcoord,chart))
                self.plotx((xcoord,chart),v=v)
